"""psutil backend."""
from typing import Callable, List, Set

import numpy as np
import psutil

from perun import config, log
from perun.backend.backend import Backend, backend
from perun.backend.device import Device
from perun.units import Byte, Percent


@backend
class PSUTIL(Backend):
    """PSUTIL Backend class."""

    name: str = "PSUTIL"
    description: str = "Obtain hardware data from psutil"

    def __init__(self) -> None:
        """Create psutil backend."""
        super().__init__()
        log.info("Init PSUTIL")

    def setup(self):
        """Configure psutil backend."""
        self.total_ram = psutil.virtual_memory().total / 1024.0**3  # byte
        self.conversion_factor = config.getfloat("devices", "ram2watt")

    def close(self):
        """Close backend."""
        pass

    def visibleDevices(self) -> Set[str]:
        """Return list of visible devices."""
        return {
            "RAM_USAGE",
            "CPU_USAGE",
            "DISK_READ_BYTES",
            "DISK_WRITE_BYTES",
        }

    def _getCallback(self, device: str) -> Callable[[], np.number]:
        """Return measuring function for each device."""
        if device == "RAM_USAGE":

            def func() -> np.number:
                return np.float32(psutil.virtual_memory().percent / 100)

        elif device == "CPU_USAGE":

            def func() -> np.number:
                return np.float64(psutil.cpu_percent())

        elif device == "DISK_READ_BYTES":

            def func() -> np.number:
                return np.uint64(psutil.disk_io_counters().read_bytes)

        elif device == "DISK_WRITE_BYTES":

            def func() -> np.number:
                return np.uint64(psutil.disk_io_counters().write_bytes)

        else:
            raise ValueError("Invalid device name")

        return func

    def getDevices(self, deviceList: Set[str]) -> List[Device]:
        """Return desired device objects."""
        for device in deviceList:
            if device == "RAM_USAGE":
                self.devices.append(
                    Device(
                        device,
                        f"{device}_psutil",
                        Percent(),
                        "",
                        np.float32(0),
                        np.float32(1.0),
                        "float32",
                        self._getCallback(device),
                    )
                )
            elif device == "CPU_USAGE":
                self.devices.append(
                    Device(
                        device,
                        f"{device}_psutil",
                        Percent(),
                        "",
                        np.float32(0),
                        np.float32(1.0),
                        "float32",
                        self._getCallback(device),
                    )
                )
            elif device == "DISK_READ_BYTES":
                self.devices.append(
                    Device(
                        device,
                        f"{device}_psutil",
                        Byte(),
                        "",
                        np.uint64(0),
                        np.uint64(np.iinfo("uint64").max),
                        "uint64",
                        self._getCallback(device),
                    )
                )
            elif device == "DISK_WRITE_BYTES":
                self.devices.append(
                    Device(
                        device,
                        f"{device}_psutil",
                        Byte(),
                        "",
                        np.uint64(0),
                        np.uint64(np.iinfo("uint64").max),
                        "uint64",
                        self._getCallback(device),
                    )
                )
        return self.devices


PSUTIL()



"""Device module."""
from dataclasses import dataclass
from typing import Any, Callable, Dict

import numpy as np

from perun.units import Unit


@dataclass
class Device:
    """Defines a devices sensor properties."""

    id: str
    long_name: str
    unit: Unit
    mag: str
    min: np.number
    max: np.number
    dtype: str
    measureCallback: Callable[[], np.number]

    def read(self) -> np.number:
        """Read value from sensor."""
        return self.measureCallback()

    def toDict(self) -> Dict[str, Any]:
        """Return device as a dictionary."""
        return {
            "id": self.id,
            "long_name": self.long_name,
            "unit": self.unit,
            "mag": self.mag,
            "min": self.min,
            "max": self.max,
            "dtype": self.dtype,
        }



"""Unit module."""
from perun import log


class MagnitudePrefix:
    """Quick conversion from metric symbols and prefixes."""

    symbols = ["p", "n", "µ", "m", "", "k", "M", "G", "T"]
    prefixes = ["pico", "nano", "micro", "mili", "", "kilo", "mega", "giga", "tera"]
    factor = [1e-12, 1e-9, 1e-6, 1e-3, 1, 1e3, 1e6, 1e9, 1e12]

    @classmethod
    def getFactor(cls, prefixOrSymbol: str) -> float:
        """Return magnitude factor based on prefix/symbol string.

        Args:
            prefixOrSymbol (str): Either a metrix prefix (ex. "mili", "giga") or symbol (ex. "m", "G")

        Raises:
            ValueError: In case prefix/symbol does not exist

        Returns:
            float: Magnitude factor as a float
        """
        if prefixOrSymbol in cls.symbols:
            idx = cls.symbols.index(prefixOrSymbol)
            return cls.factor[idx]
        elif prefixOrSymbol in cls.prefixes:
            idx = cls.prefixes.index(prefixOrSymbol)
            return cls.factor[idx]
        else:
            log.error("Invalid magnitude prefix")
            raise ValueError(prefixOrSymbol)

    @classmethod
    def getSymbol(cls, prefix: str) -> str:
        """Return symbol based on metric prefix.

        Args:
            prefix (str): Metrix prefix (ex. "mili", "tera")

        Raises:
            ValueError: In case prefix/symbol does not exist

        Returns:
            str: Metric symbol
        """
        if prefix in cls.prefixes:
            idx = cls.prefixes.index(prefix)
            return cls.symbols[idx]
        else:
            log.error("Invalid magniture prefix")
            raise ValueError(prefix)

    @classmethod
    def transformFactor(cls, fromPrefix: str, toPrefix: str) -> float:
        """
        Obtain multiplication factor to transform unit between 2 prefixes.

        Args:
            fromPrefix (str): Current magnitude prefix
            toPrefix (str): Desired magnitude prefix

        Returns:
            float: Transformation factor
        """
        fromFactor = cls.getFactor(fromPrefix)
        toFactor = cls.getFactor(toPrefix)
        return fromFactor / toFactor


class Unit:
    """Metrix unit attributes and helper methods."""

    symbol: str = "#"
    name: str = "Scalar"
    printFactor: float = 1.0

    @classmethod
    def toString(cls, value: float, mag: str = None):
        """Return string value with magnitude prefix and unit symbol."""
        magSymbol = MagnitudePrefix.getSymbol(mag) if mag else ""
        return f"{value * cls.printFactor:.3f}{magSymbol}{cls.symbol}"


class Byte(Unit):
    """Byte unit class."""

    symbol: str = "B"
    name: str = "byte"


class Percent(Unit):
    """Percent unit class."""

    symbol: str = "%"
    name: str = "Percent"
    printFactor: float = 100.0


class Joule(Unit):
    """Joule unit class."""

    symbol: str = "J"
    name: str = "Joule"

    KWH_factor = 3.6 * 1e6

    @classmethod
    def to_kWh(cls, value: float) -> float:
        """Transform from Joules to kWh."""
        return value / cls.KWH_factor

    @classmethod
    def from_kWh(cls, kwh: float) -> float:
        """Transform from kWh to Joules."""
        return kwh * cls.KWH_factor


class Watt(Unit):
    """Watt unit class."""

    symbol: str = "W"
    name: str = "Watt"


class Second(Unit):
    """Second unit class."""

    symbol: str = "s"
    name: str = "Second"



"""Configuration module."""

# Watt/GB (from : https://www.crucial.com/support/articles-faq-memory/how-much-power-does-memory-use)
# kgCO2eq/kWh - source https://www.nowtricity.com/country/germany/
# cent/kWh (Euro) - source: https://www.stromauskunft.de/strompreise/ Baden-Württemberg lokare anbieter

import configparser
from pathlib import Path
from typing import Any, Mapping, Union

import click

_default_config: Mapping[str, Mapping[str, Any]] = {
    "report": {
        "format": "txt",
        "pue": 1.58,
        "emissions-factor": 0.355,
        "price-factor": 41.59,
    },
    "devices": {
        "ram2watt": 3 / 8.0,
    },
    "monitor": {"frequency": 1, "data_out": "./results"},
    "perun": {"log_lvl": "WARN"},
    "horeka": {"enabled": False, "url": "", "token": "", "org": ""},
}

config: configparser.ConfigParser = configparser.ConfigParser()
config.read_dict(_default_config)

globalConfigPath = Path.home() / ".config/perun.ini"
if globalConfigPath.exists() and globalConfigPath.is_file():
    config.read(globalConfigPath)


def read_custom_config(
    ctx: click.Context, param: Union[click.Option, click.Parameter], pathStr: str
) -> None:
    """
    Read an INI configuration file and overrides the values from the default and global configuration.

    Args:
        ctx (click.Context): Commandline context object (irrelevant)
        param (Union[click.Option, click.Parameter]): Click cli object (irrelevant)
        pathStr (str): String to configuration file
    """
    configPath: Path = Path(pathStr)
    if configPath.exists() and configPath.is_file():
        config.read(configPath)


def save_to_config_callback(
    ctx: click.Context, param: Union[click.Option, click.Parameter], value: Any
):
    """
    Override configuration with click cli options.

    Args:
        ctx (click.Context): Click context
        param (Union[click.Option, click.Parameter]): Click option/param object
        value (Any): New configuration value
    """
    if value and isinstance(param, click.Option):
        key: str = param.name
        save_to_config(key, value)


def save_to_config(key: str, value: Any):
    """
    Override indivial configuration values.

    Args:
        key (str): Option name
        value (Any): New option value
    """
    for section in config.sections():
        if config.has_option(section, key):
            config.set(section, key, str(value))



"""Command line api definition.

Uses click https://click.palletsprojects.com/en/8.1.x/ to manage complex cmdline configurations.
"""
from datetime import datetime
from pathlib import Path
from typing import List, Optional

import click

import perun
from perun import log
from perun.configuration import config, read_custom_config, save_to_config_callback


@click.group()
@click.version_option(version=perun.__version__)
@click.option(
    "-c",
    "--configuration",
    default="./.perun.ini",
    type=click.Path(exists=False, dir_okay=False, readable=True),
    is_eager=True,
    callback=read_custom_config,
    expose_value=False,
)
@click.option(
    "-f",
    "--format",
    type=click.Choice(["txt", "yaml", "yml", "json"]),
    help="report print format",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "-f",
    "--frequency",
    type=float,
    help="sampling frequency (in Hz)",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--format",
    type=click.Choice(["txt", "yaml", "yml", "json"]),
    help="report print format",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--data_out",
    type=click.Path(exists=False, dir_okay=True, file_okay=False),
    help="experiment data output directory",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "-l",
    "--log_lvl",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"]),
    help="Loggging level",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--pue",
    type=float,
    help="Data center Power usage efficiency",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--emissions-factor",
    type=float,
    help="Emissions factor at compute resource location",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--price-factor",
    type=float,
    help="Electricity price factor at compute resource location",
    callback=save_to_config_callback,
    expose_value=False,
)
def cli():
    """Perun: Energy measuring and reporting tool."""
    log.setLevel(config.get("perun", "log_lvl"))


@cli.command()
@click.option(
    "--default",
    is_flag=True,
    show_default=True,
    default=False,
    help="Print default configuration",
)
def showconf(default: bool):
    """Print current perun configuration in INI format."""
    import sys

    from perun.configuration import _default_config

    if default:
        config.read_dict(_default_config)
        config.write(sys.stdout)
    else:
        config.write(sys.stdout)


@cli.command()
@click.argument("exp_hdf5", type=click.Path(exists=True))
def postprocess(exp_hdf5: str):
    """
    Apply post-processing to EXP_HDF5 experiment file.

    EXP_HDF5 is an hdf5 file generated by perun after monitoring a script, containing data gathered from hardware devices.
    """
    from mpi4py import MPI

    expPath = Path(exp_hdf5)
    expStrg = perun.ExperimentStorage(expPath, MPI.COMM_WORLD, write=True)
    perun.postprocessing(expStrg, reset=True)
    expStrg.close()


@cli.command()
@click.argument("exp_hdf5", type=click.Path(exists=True))
def report(exp_hdf5: str):
    """Print consumption report from EXP_HDF5 on the command line on the desired format.

    EXP_HDF5 is an hdf5 file generated by perun after monitoring a script, containing data gathered from hardware devices.
    """
    from mpi4py import MPI

    expPath = Path(exp_hdf5)
    expStrg = perun.ExperimentStorage(expPath, MPI.COMM_WORLD)
    print(perun.report(expStrg, format=config.get("report", "format")))
    expStrg.close()


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("script", type=click.Path(exists=True))
@click.argument("script_args", nargs=-1)
def monitor(
    script: str,
    script_args: tuple,
):
    """
    Gather power consumption from hardware devices while SCRIPT [SCRIPT_ARGS] is running.

    SCRIPT is a path to the python script to monitor, run with arguments SCRIPT_ARGS.
    """
    import sys
    from multiprocessing import Event, Process, Queue

    from mpi4py import MPI

    from perun import log
    from perun.backend import backends

    comm = MPI.COMM_WORLD
    start_event = Event()
    stop_event = Event()

    # Setup script arguments
    filePath: Path = Path(script)
    outPath: Path = Path(config.get("monitor", "data_out"))
    log.debug(f"Script path: {filePath}")
    argIndex = sys.argv.index(str(filePath))
    sys.argv = sys.argv[argIndex:]
    log.debug(f"Script args: { sys.argv }")

    # Get node devices
    log.debug(f"Backends: {backends}")
    lDeviceIds: List[str] = perun.getDeviceConfiguration(comm, backends)

    for backend in backends:
        backend.close()

    log.debug(f"Rank {comm.rank} - lDeviceIds : {lDeviceIds}")

    # If assigned devices, start subprocess
    if len(lDeviceIds) > 0:
        queue: Queue = Queue()
        perunSP = Process(
            target=perun.perunSubprocess,
            args=[
                queue,
                start_event,
                stop_event,
                lDeviceIds,
                config.getfloat("monitor", "frequency"),
            ],
        )
        perunSP.start()
        start_event.wait()

    # Sync everyone
    comm.barrier()
    start = datetime.now()

    # Start script
    try:
        with open(str(filePath), "r") as scriptFile:
            exec(scriptFile.read(), {"__name__": "__main__"})
    except Exception as e:
        log.error("Failed to open file ", filePath)
        log.error(e)
        stop_event.set()
        return

    stop_event.set()
    log.debug("Set closed event")
    stop = datetime.now()

    lStrg: Optional[perun.LocalStorage]
    # Obtain perun subprocess results
    if len(lDeviceIds) > 0:
        log.debug("Getting queue contents")
        lStrg = queue.get(block=True)
        log.debug("Got queue contents")
        log.debug("Waiting for subprocess to close")
        perunSP.join()
        perunSP.close()
        log.debug("Subprocess closed")
        queue.close()
    else:
        lStrg = None

    # Sync
    comm.barrier()
    log.debug("Passed first barrier")

    # Save raw data to hdf5

    if comm.rank == 0:
        if not outPath.exists():
            outPath.mkdir(parents=True)

    scriptName = filePath.name.replace(filePath.suffix, "")
    resultPath = outPath / f"{scriptName}.hdf5"
    log.debug(f"Result path: {resultPath}")
    expStrg = perun.ExperimentStorage(resultPath, comm, write=True)
    expId = expStrg.addExperimentRun(lStrg)
    if lStrg and config.getboolean("horeka", "enabled"):
        from perun.extras.horeka import get_horeka_measurements

        get_horeka_measurements(
            comm, outPath, expStrg.experimentName, expId, start, stop
        )

    # Post post-process
    comm.barrier()
    perun.postprocessing(expStorage=expStrg)
    if comm.rank == 0:
        print(
            perun.report(expStrg, expIdx=expId, format=config.get("report", "format"))
        )
    comm.barrier()
    expStrg.close()


def main():
    """Cli entrypoint."""
    cli(auto_envvar_prefix="PERUN")



"""Core perun functionality."""
import functools
import platform
import sys
import time
from datetime import datetime
from functools import reduce
from multiprocessing import Event, Process, Queue
from pathlib import Path
from typing import List, Optional, Set

import h5py
import numpy as np
from mpi4py import MPI

from perun import log
from perun.backend import Backend, Device, backends
from perun.configuration import config, read_custom_config, save_to_config
from perun.report import report
from perun.storage import ExperimentStorage, LocalStorage
from perun.units import Joule, MagnitudePrefix


def getDeviceConfiguration(comm: MPI.Comm, backends: List[Backend]) -> List[str]:
    """
    Obtain a list with the assigned devices to the current rank.

    Args:
        comm (Comm): MPI communicator object
        backends (List[Backend]): List of available backend objects

    Returns:
        List[str]: IDs of assigned devices to the current rank
    """
    visibleDevices: Set[str] = reduce(
        lambda a, b: a | b, [backend.visibleDevices() for backend in backends]
    )

    log.debug(f"Rank {comm.rank} : Visible devices {visibleDevices}")
    globalDeviceNames = comm.allgather(visibleDevices)
    globalHostnames = comm.allgather(platform.node())

    previousHosts = {}
    for index, (hostname, globalDevices) in enumerate(
        zip(globalHostnames, globalDeviceNames)
    ):
        if hostname not in previousHosts:
            previousHosts[hostname] = index
        else:
            prevIndex = previousHosts[hostname]
            globalDeviceNames[prevIndex] |= globalDevices
            globalDeviceNames[index] = set()

    return globalDeviceNames[comm.rank]


def monitor(
    configuration: str = "./.perun.ini",
    **conf_kwargs,
):
    """Decorate function to monitor its energy usage."""

    def inner_function(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):

            # Get custom config and kwargs
            read_custom_config(None, None, configuration)
            for key, value in conf_kwargs.items():
                save_to_config(key, value)

            comm = MPI.COMM_WORLD
            start_event = Event()
            stop_event = Event()

            filePath: Path = Path(sys.argv[0])
            outPath: Path = Path(config.get("monitor", "data_out"))

            # Get node devices
            log.debug(f"Backends: {backends}")
            lDeviceIds: List[str] = getDeviceConfiguration(comm, backends)

            for backend in backends:
                backend.close()

            log.debug(f"Rank {comm.rank} - lDeviceIds : {lDeviceIds}")

            # If assigned devices, start subprocess
            if len(lDeviceIds) > 0:
                queue: Queue = Queue()
                perunSP = Process(
                    target=perunSubprocess,
                    args=[
                        queue,
                        start_event,
                        stop_event,
                        lDeviceIds,
                        config.getfloat("monitor", "frequency"),
                    ],
                )
                perunSP.start()
                start_event.wait()

            # Sync everyone
            comm.barrier()
            start = datetime.now()

            func(*args, **kwargs)
            stop_event.set()
            stop = datetime.now()

            lStrg: Optional[LocalStorage]
            # Obtain perun subprocess results
            if len(lDeviceIds) > 0:
                perunSP.join()
                perunSP.close()
                lStrg = queue.get()
            else:
                lStrg = None

            # Sync
            comm.barrier()

            # Save raw data to hdf5

            if comm.rank == 0:
                if not outPath.exists():
                    outPath.mkdir(parents=True)

            scriptName = filePath.name.replace(filePath.suffix, "")
            resultPath = outPath / f"{scriptName}.hdf5"
            log.debug(f"Result path: {resultPath}")
            expStrg = ExperimentStorage(resultPath, comm, write=True)
            expId = expStrg.addExperimentRun(lStrg)
            if lStrg and config.getboolean("horeka", "enabled"):
                from perun.extras.horeka import get_horeka_measurements

                get_horeka_measurements(
                    comm, outPath, expStrg.experimentName, expId, start, stop
                )

            # Post post-process
            comm.barrier()
            postprocessing(expStorage=expStrg)
            if comm.rank == 0:
                print(
                    report(expStrg, expIdx=expId, format=config.get("report", "format"))
                )
            comm.barrier()
            expStrg.close()

        return func_wrapper

    return inner_function


def perunSubprocess(
    queue: Queue, start_event, stop_event, deviceIds: Set[str], frequency: float
):
    """
    Parallel function that samples energy values from hardware libraries.

    Args:
        queue (Queue): multiprocessing Queue object where the results are sent after finish
        start_event (_type_): Marks that perun finished setting up and started sampling from devices
        stop_event (_type_): Signal the subprocess that the monitored processed has finished
        deviceIds (Set[str]): List of device ids to sample from
        frequency (int): Sampling frequency in Hz
    """
    from perun.backend import backends

    lDevices: List[Device] = []
    for backend in backends:
        backend.setup()
        lDevices += backend.getDevices(deviceIds)

    log.debug(f"Perun SP lDevices: {lDevices}")
    lStrg = LocalStorage(platform.node(), lDevices)
    start_event.set()

    while not stop_event.wait(1.0 / frequency):
        t = np.uint64(time.time_ns())
        stepData = {}
        for device in lDevices:
            stepData[device.id] = device.read()
        lStrg.addTimestep(t, stepData)

    log.debug("Subprocess: Stop event received.")
    for backend in backends:
        backend.close()
    log.debug("Subprocess: Closed backends")

    queue.put(lStrg, block=True)
    log.debug("Subprocess: Sent lStrg")


def postprocessing(expStorage: ExperimentStorage, reset: bool = False):
    """Process the obtained data."""
    if (expStorage.serial and expStorage.comm.rank == 0) or not expStorage.serial:
        expRuns = expStorage.getExperimentRuns()
        totalExpEnergy_kWh = 0
        totalExpRuntime_s = 0
        totalExpCO2e_kg = 0
        totalExpPrice_euro = 0
        for run in expRuns:
            if "totalRunEnergy_kWh" not in run.attrs or reset:
                _postprocessRun(run, reset)
            totalExpEnergy_kWh += run.attrs["totalRunEnergy_kWh"]
            totalExpRuntime_s += run.attrs["totalRunRuntime_s"]
            totalExpCO2e_kg += run.attrs["totalRunCO2e_kg"]
            totalExpPrice_euro += run.attrs["totalRunPrice_euro"]

        rootExp = expStorage.getRootObject()
        rootExp.attrs["totalExpEnergy_kWh"] = totalExpEnergy_kWh
        rootExp.attrs["totalExpRuntime_s"] = totalExpRuntime_s
        rootExp.attrs["totalExpCO2e_kg"] = totalExpCO2e_kg
        rootExp.attrs["totalExpPrice_euro"] = totalExpPrice_euro


def _postprocessRun(run: h5py.Group, reset: bool = False):
    """
    Collect run statistics and save them in the group.

    Args:
        run (h5py.Group): Run hdf5 group
    """
    totalRunEnergy_kJ = 0
    totalRunRuntime_s = 0
    runAvgPower_kW = 0
    totalRunRuntime_s = 0

    for node in run.values():
        if "totalNodeEnergy_kJ" not in node.attrs or reset:
            _postprocessNode(node, reset)
        totalRunEnergy_kJ += node.attrs["totalNodeEnergy_kJ"]
        runAvgPower_kW += node.attrs["avgNodePower_kW"]
        totalRunRuntime_s = max(totalRunRuntime_s, node.attrs["totalNodeRuntime_s"])

    totalRunEnergy_kWh = Joule.to_kWh(
        totalRunEnergy_kJ * MagnitudePrefix.transformFactor("kilo", "")
    ) * config.getfloat("report", "pue")
    run.attrs["totalRunEnergy_kWh"] = totalRunEnergy_kWh
    run.attrs["totalRunRuntime_s"] = totalRunRuntime_s
    run.attrs["runAvgPower_kW"] = runAvgPower_kW

    run.attrs["totalRunCO2e_kg"] = totalRunEnergy_kWh * config.getfloat(
        "report", "emissions-factor"
    )
    run.attrs["totalRunPrice_euro"] = (
        totalRunEnergy_kWh * config.getfloat("report", "price-factor") / 100
    )


def _postprocessNode(node: h5py.Group, reset: bool = False):
    """
    Collect node statistics and save them in the group.

    Args:
        node (h5py.Group): Node hdf5 group
    """
    totalNodeEnergy_kJ = 0
    avgNodePower_kW = 0
    t_ns_ds: h5py.Dataset = node["t_ns"]

    mag = t_ns_ds.attrs["mag"]
    transformFactor = MagnitudePrefix.transformFactor(mag, "")
    node.attrs["totalNodeRuntime_s"] = (t_ns_ds[-1] - t_ns_ds[0]) * transformFactor

    for device in node.values():
        if "/t_ns" not in device.name and device.attrs["units"] in ["Joule", "Watt"]:
            if "totalDeviceEnergy_kJ" not in device.attrs or reset:
                _postprocessDevice(device, t_ns_ds)
            totalNodeEnergy_kJ += device.attrs["totalDeviceEnergy_kJ"]
            avgNodePower_kW += device.attrs["avgDevicePower_kW"]

    node.attrs["totalNodeEnergy_kJ"] = totalNodeEnergy_kJ
    node.attrs["avgNodePower_kW"] = avgNodePower_kW


def _postprocessDevice(device: h5py.Dataset, t_ns: h5py.Dataset):
    """
    Process device raw data and gather report statistics.

    Args:
        device (h5py.Dataset): Device hdf5 dataset
        t_ns (h5py.Dataset): Timesteps hdf5 dataset
    """
    mag = device.attrs["scale_prefix"]
    mTransFactor = MagnitudePrefix.transformFactor(mag, "kilo")
    tTransFactor = MagnitudePrefix.transformFactor(t_ns.attrs["mag"], "")

    if device.attrs["units"] == "Joule":
        energy_array = device[:]
        maxValue = device.attrs["valid_max"]
        dtype = device.dtype.name
        d_energy = energy_array[1:] - energy_array[:-1]
        if "uint" in dtype:
            idx = d_energy >= maxValue
            max_dtype = np.iinfo(dtype).max
            d_energy[idx] = maxValue + d_energy[idx] - max_dtype
        else:
            idx = d_energy <= 0
            d_energy[idx] = d_energy[idx] + maxValue
        total_energy = d_energy.sum()

        totalDeviceEnergy_kJ = float(total_energy) * mTransFactor
        device.attrs["totalDeviceEnergy_kJ"] = totalDeviceEnergy_kJ
        device.attrs["avgDevicePower_kW"] = totalDeviceEnergy_kJ / (
            (t_ns[-1] - t_ns[0]) * tTransFactor
        )
    elif device.attrs["units"] == "Watt":
        power_w = device[:] * mTransFactor
        t_s = t_ns[:] * tTransFactor
        energy_kj = np.trapz(power_w, t_s)
        device.attrs["totalDeviceEnergy_kJ"] = energy_kj
        device.attrs["avgDevicePower_kW"] = np.mean(power_w)
